/* usbmsgd: Usb mass storage gadget daemon to access usb storage
 * Copyright (C) 2007-2011 David Zeuthen <zeuthen@gmail.com>
 * Copyright (C) 2018 Robert Bosch Power Tools GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef __USBMSGD_LOGGING_H__
#define __USBMSGD_LOGGING_H__

#include <gio/gio.h>
G_BEGIN_DECLS


/**
 * USBMSGDLogLevel:
 * @USBMSGD_LOG_LEVEL_DEBUG: Debug messages.
 * @USBMSGD_LOG_LEVEL_INFO: Informational messages.
 * @USBMSGD_LOG_LEVEL_NOTICE: Messages that the administrator should take notice of.
 * @USBMSGD_LOG_LEVEL_WARNING: Warning messages.
 * @USBMSGD_LOG_LEVEL_ERROR: Error messages.
 *
 * Logging levels. The level @USBMSGD_LOG_LEVEL_NOTICE and above goes to journal.
 *
 * Unlike g_warning() and g_error(), none of these logging levels causes the program to ever terminate.
 */
typedef enum
{
  USBMSGD_LOG_LEVEL_DEBUG = G_LOG_LEVEL_DEBUG,
  USBMSGD_LOG_LEVEL_INFO = G_LOG_LEVEL_INFO,
  USBMSGD_LOG_LEVEL_MESSAGE = G_LOG_LEVEL_MESSAGE,
  USBMSGD_LOG_LEVEL_WARNING = G_LOG_LEVEL_WARNING,
  USBMSGD_LOG_LEVEL_CRITICAL = G_LOG_LEVEL_CRITICAL,
  USBMSGD_LOG_LEVEL_ERROR = G_LOG_LEVEL_ERROR
} UsbmsgdLogLevel;


void usbmsgd_log (UsbmsgdLogLevel   level,
                 const gchar     *function,
                 const gchar     *location,
                 const gchar     *format,
                 ...) G_GNUC_PRINTF (4, 5);

/**
 * usbmsgd_debug:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %USBMSGD_LOG_LEVEL_DEBUG.
 * Warning: ONLY WORKS WHEN COMPILED WITH --enable-debug
 *
 * See #UsbmsgdLogLevel for more details.
 */
#ifdef DEBUG
#define usbmsgd_debug(args...)   usbmsgd_log(USBMSGD_LOG_LEVEL_DEBUG, G_STRFUNC, G_STRLOC, args)
#else
#define usbmsgd_debug(args...)
#endif

/**
 * usbmsgd_info:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %USBMSGD_LOG_LEVEL_INFO.
 * Warning: ONLY WORKS WHEN COMPILED WITH --enable-debug
 *
 * See #UsbmsgdLogLevel for more details.
 */
#ifdef DEBUG
#define usbmsgd_info(args...)    usbmsgd_log(USBMSGD_LOG_LEVEL_INFO, G_STRFUNC, G_STRLOC, args)
#else
#define usbmsgd_info(args...)
#endif

/**
 * usbmsgd_notice:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %USBMSGD_LOG_LEVEL_MESSAGE.
 *
 * See #UsbmsgdLogLevel for more details.
 */
#define usbmsgd_notice(args...)    usbmsgd_log(USBMSGD_LOG_LEVEL_MESSAGE, G_STRFUNC, G_STRLOC, args)

/**
 * usbmsgd_message:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %USBMSGD_LOG_LEVEL_MESSAGE.
 *
 * See #UsbmsgdLogLevel for more details.
 */
#define usbmsgd_message(args...)    usbmsgd_log(USBMSGD_LOG_LEVEL_MESSAGE, G_STRFUNC, G_STRLOC, args)

/**
 * usbmsgd_warning:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %USBMSGD_LOG_LEVEL_WARNING.
 *
 * See #UsbmsgdLogLevel for more details.
 */
#define usbmsgd_warning(args...) usbmsgd_log(USBMSGD_LOG_LEVEL_WARNING, G_STRFUNC, G_STRLOC, args)

/**
 * usbmsgd_critical:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %USBMSGD_LOG_LEVEL_CRITICAL.
 *
 * See #UsbmsgdLogLevel for more details.
 */
#define usbmsgd_critical(args...)   usbmsgd_log(USBMSGD_LOG_LEVEL_CRITICAL, G_STRFUNC, G_STRLOC, args)

/**
 * usbmsgd_error:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %USBMSGD_LOG_LEVEL_ERROR.
 * Warning: THIS ALSO TERMINATES THE PROCESS with the abort() call!
 *
 * See #UsbmsgdLogLevel for more details.
 */
#define usbmsgd_error(args...)   usbmsgd_log(USBMSGD_LOG_LEVEL_ERROR, G_STRFUNC, G_STRLOC, args)


G_END_DECLS

#endif /* __USBMSGD_LOGGING_H__ */
