/* usbmsgd: Usb mass storage gadget daemon to access usb storage
 * Copyright (C) 2018 Robert Bosch Power Tools GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Author: Vinay Kumar Shettar <vinaykumar.shettar@in.bosch.com>
 */
/**
 * @file main.cpp
 * @brief the main code of the laser range finder demon
 * @mainpage usbmsgd main purpose:
 * The purpose of the usbmsgd - Usb mass storage gadget daemon - is to
 * bring the usb mass Storage gadget daemon module into the d-bus world.\n
 * The GUI will be using it via d-bus then. So the usbmsgd is mainly
 * a d-bus proxy for the mass storage functionality (this will be shown on the
 * main documentation page)\n For details on D-bus concept, see:
 * https://dbus.freedesktop.org/doc/dbus-tutorial.html \n For Details on
 * GMainLoop Concepts, see:
 * https://developer.gnome.org/glib/stable/glib-The-Main-Event-Loop.html
 */
#include "usbmsgd-logging.h"
#include <glib-unix.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <signal.h>
#include <usbmsgd.h>

/**
 * @author Vinay Kumar
 * @date 28 May 2020
 * @brief callback function to handle unix quit signal, it makes
 * the GMainLoop to quit.
 * @param user_data user data.
 * @return TRUE for continue to recive signals FALSE otherwises.
 */
static gboolean
handle_quit_signal (gpointer user_data)
{
  usbmsgd_debug ("Caught Quit Signal\n");

  GMainLoop *ploop = (GMainLoop *)user_data;
  // quit GMainLoop.
  if (ploop)
    g_main_loop_quit (ploop);

  return FALSE;
}

/**
 * @author Vinay Kumar
 * @date 28 May 2020
 * @brief  callback to handle unix reload signal.
 * @param user_data user data.
 * @return TRUE for continue to recive signals FALSE otherwises.
 */
gboolean
handle_reload_signal (gpointer user_data)
{
  usbmsgd_debug ("Caught Service Reload Signal\n");
  return TRUE;
}

/**
 * @author Vinay Kumar
 * @date 28 May 2020
 * @brief setup mount directory for mount stroge partition in device.
 * @return TRUE on success of setup of mount dir FALSE otherwises.
 */
static gboolean
setup_mount_dir (void)
{
  if (!g_file_test (USBMSGD_MOUNT_DIR, G_FILE_TEST_IS_DIR))
    {
      int ret;
      ret = g_mkdir_with_parents (USBMSGD_MOUNT_DIR, 0766);

      if (ret != 0)
        {
          usbmsgd_critical ("Couldn't create mount directory\n");
          return FALSE;
        }
      else
        {
          g_chmod (USBMSGD_MOUNT_DIR, 0766);
          usbmsgd_debug ("Directory Created\n");
        }
    }
  else
    {
      g_chmod (USBMSGD_MOUNT_DIR, 0766);
      usbmsgd_debug ("Directory present\n");
    }
  return TRUE;
}

/**
 * @author Vinay Kumar
 * @date 28 May 2020
 * @brief Main entry point of program.
 * - Installs D-bus callbacks.
 * - Runs g_main_loop, It manages all the available sources of events for Glib
 * applications.
 * @param argc Number of command line arguments.
 * @param argv Array of command line arguments.
 */

gint
main (gint argc, gchar *argv[])
{
  gint opt;
  USBMSG *usbmsg;
  /* Initialize the gnome type system */
#ifndef GLIB_DEPRECATED_IN_2_36
  g_type_init (); /* Function is deprecated Since 2.36 the type system is now
                     initialised automatically. Not needed for SYSTEMBUS. */
#endif

  /* GMainLoop struct is an opaque data type representing the main event loop
   *  of a GLib application
   */
  guint bus_name_id;

  GBusType BusType = G_BUS_TYPE_SYSTEM;
  usbmsgd_debug ("SYSTEMBUS\n");

  if (!setup_mount_dir ())
    {
      usbmsgd_critical ("setup_mount_dir failed\n");
    }

  usbmsg = g_new0 (USBMSG, 1);

  /* Create gnome main loop with default context, not running yet */
  usbmsg->loop = g_main_loop_new (NULL, FALSE);

  /*
   * Init D-Bus
   * Make the deamon known on the D-bus.
   * Callbacks will be invoked in the thread-default main context of the thread
   * you are calling this function from.
   */
  bus_name_id = g_bus_own_name (BusType, "com.bosch_pt.Usbmsgd",
                                G_BUS_NAME_OWNER_FLAGS_NONE,
                                usbmsgd_bus_acquired, usbmsgd_name_acquired,
                                usbmsgd_name_lost, usbmsg, NULL);
  usbmsgd_debug ("Real Server started\n");

  /* Install Unix Signal Handler.*/
  g_unix_signal_add (SIGINT, handle_quit_signal, usbmsg->loop);
  g_unix_signal_add (SIGTERM, handle_quit_signal, usbmsg->loop);
  g_unix_signal_add (SIGHUP, handle_reload_signal, usbmsg->loop);

  /* Run the main loop until g_main_loop_quit() is called on the loop.*/
  usbmsgd_debug ("Running main loop...\n");
  g_main_loop_run (usbmsg->loop);

  usbmsgd_debug ("Left main loop\n");

  usbmsgd_deinit (usbmsg);

  /* Clean up, Remove name from D-bus */
  g_bus_unown_name (bus_name_id);

  if(usbmsg->usbmsgd_object)
       g_object_unref(usbmsg->usbmsgd_object);
  /* Clean up gnome stuff */
  g_main_loop_unref (usbmsg->loop);
  g_free (usbmsg);
  return 0;
}
