/* usbmsgd: Usb mass storage gadget daemon to access usb storage
 * Copyright (C) 2018 Robert Bosch Power Tools GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Author: Vinay Kumar Shettar <vinaykumar.shettar@in.bosch.com>
 */

#include "usbmsgd-logging.h"
#include <blockdev/blockdev.h>
#include <blockdev/fs.h>
#include <blockdev/loop.h>
#include <errno.h>
#include <fcntl.h>
#include <gudev/gudev.h>
#include <linux/usb/ch9.h>
#include <usbg/function/ms.h>
#include <usbg/usbg.h>
#include <usbmsgd.h>

static usbg_state *s;
static usbg_gadget *g;

#define USBMSGD_GADGET_CONFIG_FILE                                            \
  "/sys/kernel/config/usb_gadget/g1/functions/mass_storage.0/lun.0/file"

static gchar *ms_pc_usb_port      = NULL;
static GIOChannel *ms_usb_channel = NULL;
static guint ms_watch_id          = 0;

typedef struct
{
  const gchar *subsystem;
  const gchar *driver;
} DeviceInfo;

#define DEVICE_COUNT 1
static DeviceInfo device[DEVICE_COUNT] = {
  // am335x usb mass storage
  {
      .subsystem = "platform",
      .driver    = "musb-hdrc",
  },
};

typedef struct
{
  USBMSG *usbmsg;
  gchar *label;
} format_data;

/**
 * @brief match the device form subsystem device list.
 * @param client GUdevClient - query devices.
 * @param subsystem_match - subsystem to get devices
 * @param driver_match - reference name of the driver used for device.
 * @return information about a device on success and NULL othewise.
 */
static GUdevDevice *
match_device (GUdevClient *client, const gchar *subsystem_match,
              const gchar *driver_match)
{
  GList *list;
  GList *list_ref;
  GUdevDevice *device_data     = NULL;
  GUdevDevice *device_data_ref = NULL;

  gboolean ret = FALSE;
  const gchar *ofnode;
  gchar *drmode = NULL;

  list = g_udev_client_query_by_subsystem (client, subsystem_match);
  for (list_ref = list; list_ref; list_ref = list_ref->next)
    {
      const gchar *driver;
      const gchar *device_otg_sysfs_path;

      device_data = (GUdevDevice *)list_ref->data;
      driver      = g_udev_device_get_driver (device_data);

      if (driver && driver_match)
        {
          if (strcmp (driver, driver_match) == 0)
            {
              device_otg_sysfs_path
                  = g_udev_device_get_sysfs_path (device_data);
              ofnode = g_strconcat (device_otg_sysfs_path, "/of_node/dr_mode",
                                    NULL);

              ret = g_file_get_contents (ofnode, &drmode, NULL, NULL);
              if (ret)
                {
                  if (strcmp (drmode, "peripheral") == 0)
                    {
                      device_data_ref = device_data;
                      g_list_free (list);
                      return device_data_ref;
                    }
                }
            }
        }
    }
  return device_data_ref;
}

/**
 * @brief find the device form devices list.
 * @param client GUdevClient - query devices.
 * @return information about a device on success and NULL othewise.
 */
static GUdevDevice *
search_device (GUdevClient *client)
{
  gint count;
  DeviceInfo *device_data;
  GUdevDevice *device_otg = NULL;

  for (count = 0; count < DEVICE_COUNT; count++)
    {
      device_otg  = NULL;
      device_data = &device[count];

      device_otg
          = match_device (client, device_data->subsystem, device_data->driver);
      if (!device_otg)
        {
          continue;
        }
      usbmsgd_debug ("Found otg\n");
      break;
    }
  return device_otg;
}

/**
 * @author Gireesh Hiremath
 * @date 25.06.2020
 * @brief Get the path of usb status.
 * @return 1 on getting otg path and 0 othewise.
 */
static int
get_usb_port ()
{
  GUdevClient *client       = NULL;
  GUdevDevice *device_otg   = NULL;
  const gchar *subsystems[] = { NULL };
  const gchar *device_otg_sysfs_path;

  client = g_udev_client_new (subsystems);
  if (!client)
    {
      usbmsgd_critical ("USB otg subsystem not found\n");
      return 0;
    }

  if ((device_otg = search_device (client)) == NULL)
    {
      usbmsgd_critical ("USB otg device not found\n");
      return 0;
    }

  device_otg_sysfs_path = g_udev_device_get_sysfs_path (device_otg);
  ms_pc_usb_port        = g_strconcat (device_otg_sysfs_path, "/mode", NULL);
  usbmsgd_debug ("ms_pc_usb_port %s\n", ms_pc_usb_port);
  return 1;
}

/**
 * @author Vinay
 * @date 27.02.2020
 * @brief Mount the Mass Storage Partition to Device.
 * @param mount_path path to mount the storage patition.
 * @return 1 on mount success and 0 othewise.
 */
static gint
mount_for_device (USBMSG *usbmsg, const gchar *mount_path,
                  const gchar *mount_option)
{
  g_autoptr (GError) error = NULL;
  gchar *fstype            = NULL;
  gint ret                 = 0;
  gchar *mount_path_exist  = NULL;

  if (usbmsg->device_name != NULL)
    {
      mount_path_exist = bd_fs_get_mountpoint (usbmsg->device_name, &error);
      if (mount_path_exist == NULL)
        {
          fstype = bd_fs_get_fstype (usbmsg->device_name, &error);
          if (fstype == NULL)
            {
              usbmsgd_critical ("Error bd_fs_get_fstype:");
              if (error)
                {
                  usbmsgd_critical ("(%s, %d)\n", error->message, error->code);
                  g_clear_error (&error);
                }
            }
          else
            {
              usbmsgd_debug ("bd_fs_get_fstype : %s\n", fstype);
            }
          /*
           * TODO: Get the mount option from client process to set uid and gid
           *      to mount with normal user permission.
           */
          ret = bd_fs_mount (usbmsg->device_name, mount_path, fstype,
                             "rw,uid=1000,gid=1000", NULL, &error);
          if (!ret)
            {
              usbmsgd_critical ("Error bd_fs_mount: %d\n", ret);
              if (error)
                {
                  usbmsgd_critical ("(%s, %d)\n", error->message, error->code);
                  g_clear_error (&error);
                }
            }
          else
            {
              usbmsgd_debug ("bd_fs_Mount Done: %d\n", ret);
              com_bosch_pt_usbmsgd_set_mount_point (usbmsg->usbmsgd_object,
                                                    mount_path);
            }
        }
      else
        {
          usbmsgd_debug ("Mount path exist: %s\n", mount_path_exist);
          com_bosch_pt_usbmsgd_set_mount_point (usbmsg->usbmsgd_object,
                                                mount_path_exist);
        }
    }
  else
    usbmsgd_critical ("Mount for device Failed\n");
  return ret;
}

/**
 * @author Vinay
 * @date 27.02.2020
 * @brief Unmount the Mass Storage Partition to Device.
 * @return 1 on mount success and 0 othewise.
 */
static gint
unmount_for_device (USBMSG *usbmsg)
{
  g_autoptr (GError) error = NULL;
  gint ret                 = 0;
  gchar *mount_path_exist  = NULL;
  /*
   * Check if storage is mounted already or not with bd_fs_get_mountpoint()
   * This is needed if device connected to PC at init and unmount is called by
   * applicaton.
   */
  if (usbmsg->device_name != NULL)
    {
      mount_path_exist = bd_fs_get_mountpoint (usbmsg->device_name, &error);
      if (mount_path_exist != NULL)
        {
          ret = bd_fs_unmount (usbmsg->device_name, 1, 0, NULL, &error);
          if (!ret)
            {
              usbmsgd_critical ("Error bd_fs_unmount: %d\n", ret);
              if (error)
                usbmsgd_critical ("(%s, %d)\n", error->message, error->code);
            }
          else
            {
              usbmsgd_debug ("bd_fs_unmount Done: %d\n", ret);
              com_bosch_pt_usbmsgd_set_mount_point (usbmsg->usbmsgd_object,
                                                    NULL);
            }
        }
      else
        {
          usbmsgd_debug ("mount path does not exist\n");
          com_bosch_pt_usbmsgd_set_mount_point (usbmsg->usbmsgd_object, NULL);
        }
    }
  else
    usbmsgd_critical ("unmount for device Failed\n");

  if (com_bosch_pt_usbmsgd_get_is_usb_connected (usbmsg->usbmsgd_object))
    {
      usbmsgd_set_gadget_config_file_content (USBMSGD_BACKING_FILE);
    }

  return ret;
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief Create the loop device for internal mass storage.
 */
static gint
setup_loopdev (USBMSG *usbmsg)
{
  gint ret                 = 0;
  g_autoptr (GError) error = NULL;
  const gchar *loop_name   = NULL;
  gchar *loop_name_exist   = NULL;
  /* TODO: Read backing file from /etc/usbmsgd/config.txt */
  loop_name_exist = bd_loop_get_loop_name (USBMSGD_BACKING_FILE, &error);
  if (loop_name_exist == NULL)
    {
      ret = bd_loop_setup (USBMSGD_BACKING_FILE, 0, 0, FALSE, FALSE,
                           &loop_name, &error);
      if (!ret)
        {
          usbmsgd_error ("Error bd_loop_setup: %d\n", ret);
          if (error)
            usbmsgd_critical ("Error (%s, %d)\n", error->message, error->code);
          g_clear_error (&error);
        }
      else
        {
          usbmsgd_debug ("bd_loop_setup Done: %d, Loop_name =%s\n", ret,
                         loop_name);
          usbmsg->device_name = (char *)g_malloc (5 + strlen (loop_name));
          strcpy (usbmsg->device_name, "/dev/");
          strcat (usbmsg->device_name, loop_name);
          usbmsgd_debug ("device_name : %s\n", usbmsg->device_name);
        }
    }
  else
    {
      usbmsgd_debug ("bd_loop_get_loop_name : %s\n", loop_name_exist);
      usbmsg->device_name = (char *)malloc (5 + strlen (loop_name_exist));
      strcpy (usbmsg->device_name, "/dev/");
      strcat (usbmsg->device_name, loop_name_exist);
      usbmsgd_debug ("device_name : %s\n", usbmsg->device_name);
    }
  return ret;
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief Teardownthe loop device for internal mass storage.
 */
static gint
teardown_loopdev (USBMSG *usbmsg)
{
  gint ret                 = 0;
  g_autoptr (GError) error = NULL;
  gchar *loop_name         = NULL;

  /* Unmount the device if mounted before teardown loopdev. */
  unmount_for_device (usbmsg);

  loop_name = bd_loop_get_loop_name (USBMSGD_BACKING_FILE, &error);
  if (loop_name == NULL)
    {
      usbmsgd_debug ("Error bd_loop_get_loop_name:");
      if (error)
        {
          usbmsgd_critical ("(%s, %d)\n", error->message, error->code);
          g_clear_error (&error);
        }
    }
  else
    {
      usbmsgd_debug ("bd_loop_get_loop_name : %s\n", loop_name);
    }
  ret = bd_loop_teardown (loop_name, &error);
  if (!ret)
    {
      usbmsgd_critical ("bd_loop_teardown: %d\n", ret);
      if (error)
        usbmsgd_critical ("(%s, %d)\n", error->message, error->code);
      g_clear_error (&error);
    }
  else
    {
      usbmsgd_debug ("bd_loop_teardown Done: %d\n", ret);
    }
  if (usbmsg->device_name != NULL)
    {
      g_free (usbmsg->device_name);
    }
  usbmsg->device_name = NULL;
  return ret;
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief Initilize fs and loop Plugin in libblockdev's function..
 */
static gint
blockdev_plugin_init ()
{
  gint ret;
  g_autoptr (GError) error = NULL;
  BDPluginSpec fs_plugin   = { BD_PLUGIN_FS, NULL };
  BDPluginSpec loop_plugin = { BD_PLUGIN_LOOP, NULL };
  /*
   * The core daemon needs the part, swap, loop, mdraid, fs and crypto plugins.
   *   Additional plugins are required by various modules, but they make sure
   *   plugins are loaded themselves.
   */
  BDPluginSpec *plugins[] = { &fs_plugin, &loop_plugin, NULL };
  BDPluginSpec **plugin_p = NULL;
  error                   = NULL;

  /*
   * Skip runtime dependency checks when initializing libblockdev. Plugin
   *   shouldn't fail to load just because some if its dependencies is missing.
   */
  ret = bd_switch_init_checks (FALSE, &error);
  if (!ret)
    {
      if (error)
        {
          usbmsgd_error ("Error initializing libblockdev library: (%s, %d)",
                         error->message, error->code);
          g_clear_error (&error);
        }
    }

  ret = bd_try_init (plugins, NULL, NULL, &error);
  if (!ret)
    {
      if (error)
        {
          usbmsgd_error ("Error initializing libblockdev library:(%s, %d)",
                         error->message, error->code);
          g_clear_error (&error);
        }
      else
        {
          /*
           * a missing plugin is okay, calling functions from it will fail, but
           *   until that happens, life will just be great
           */
          for (plugin_p = plugins; *plugin_p; plugin_p++)
            if (!bd_is_plugin_available ((*plugin_p)->name))
              usbmsgd_error ("Failed to load the '%s' libblockdev plugin",
                             bd_get_plugin_name ((*plugin_p)->name));
        }
    }

  usbmsgd_debug ("blockdev_plugin_init Done\n");

  return 0;
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief filesystem repair is performerd.
 * @return returns TRUE if success and FALSE otherwise.
 */
static gboolean
check_filesystem (USBMSG *usbmsg)
{
  gint ret                 = 0;
  gboolean ret_val         = FALSE;
  g_autoptr (GError) error = NULL;

  usbmsgd_debug ("check_filesystem\n");
  if (usbmsg->device_name != NULL)
    {
      ret = bd_fs_check (usbmsg->device_name, &error);
      if (!ret)
        {
          usbmsgd_debug ("Error bd_fs_check: %d\n", ret);
          if (error)
            {
              usbmsgd_debug ("%s, %d)\n", error->message, error->code);
              g_clear_error (&error);
            }
        }
      else
        {
          usbmsgd_debug ("bd_fs_check Done: %d\n", ret);
          ret_val = TRUE;
        }
    }
  else
    usbmsgd_critical ("check filesystem Failed\n");
  return ret_val;
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief filesystem check is performerd.
 * @return gboolean returns TRUE if success and FALSE otherwise.
 */
static gboolean
repair_filesystem (USBMSG *usbmsg)
{
  gint ret                 = 0;
  gboolean ret_val         = FALSE;
  g_autoptr (GError) error = NULL;

  usbmsgd_debug ("repair_filesystem\n");
  if (usbmsg->device_name != NULL)
    {
      ret = bd_fs_repair (usbmsg->device_name, &error);
      if (!ret)
        {
          usbmsgd_critical ("Error bd_fs_repair: %d\n", ret);
          if (error)
            {
              usbmsgd_critical ("Error (%s, %d)\n", error->message,
                                error->code);
            }
        }
      else
        {
          usbmsgd_debug ("bd_fs_repair Done: %d\n", ret);
          ret_val = TRUE;
        }
    }
  else
    usbmsgd_critical ("repair filesystem Failed\n");
  return ret_val;
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief filesystem check is performerd and if check failed it will perform
 * filesystem repair On failure of filesystem check and repair it will send
 * filesystem corrupted d-bus notification.
 */
static void
check_and_repair_filesystem (USBMSG *usbmsg)
{
  gchar *fstype                 = NULL;
  gint ret                      = 0;
  g_autoptr (GError) error      = NULL;
  gboolean filesystem_corrupted = FALSE;
  usbmsgd_debug ("check_and_repair_filesystem\n");

  if (usbmsg->device_name != NULL)
    {
      fstype = bd_fs_get_fstype (usbmsg->device_name, &error);
      if (fstype == NULL)
        {
          usbmsgd_critical ("Error bd_fs_get_fstype\n");
          if (error)
            {
              usbmsgd_critical ("Error (%s, %d)\n", error->message,
                                error->code);
              g_clear_error (&error);
            }
          filesystem_corrupted = TRUE;
          com_bosch_pt_usbmsgd_set_file_system_type (usbmsg->usbmsgd_object,
                                                     "unknown");
        }
      else
        {
          usbmsgd_debug ("bd_fs_get_fstype Done: %s\n", fstype);
          ret = check_filesystem (usbmsg);
          if (!ret)
            {
              /* File system is not ok so try to repair the same. */
              ret = repair_filesystem (usbmsg);
              if (!ret)
                {
                  filesystem_corrupted = TRUE;
                }
              else
                {
                  usbmsgd_debug ("repair_filesystem done\n");
                }
            }
          else
            {
              usbmsgd_debug ("check_filesystem done\n");
            }
          com_bosch_pt_usbmsgd_set_file_system_type (usbmsg->usbmsgd_object,
                                                     fstype);
        }

      if (filesystem_corrupted == TRUE)
        {

          com_bosch_pt_usbmsgd_set_is_file_system_corrupted (
              usbmsg->usbmsgd_object, TRUE);
          com_bosch_pt_usbmsgd_set_format_status (
              usbmsg->usbmsgd_object, USBMSGD_FORMAT_STATE_UNKNOWN);
        }
      else
        {
          com_bosch_pt_usbmsgd_set_is_file_system_corrupted (
              usbmsg->usbmsgd_object, FALSE);
        }
    }
  else
    usbmsgd_critical ("check and repair filesystem Failed\n");
}

/**
 * @author Vinay
 * @date 10.06.2022
 * @brief usbmsgd_set_gadget_config_file_content
 * @param str the gadget file path
 * @return TRUE on success, FALSE if an error occurred.
 */
int
usbmsgd_set_gadget_config_file_content (const char *str)
{
  int fd;
  fd = open (USBMSGD_GADGET_CONFIG_FILE, O_WRONLY | O_TRUNC);

  if (fd < 0)
    {
      usbmsgd_warning ("Error opening gadget config file\n");
      return 1;
    }
  write (fd, str, strlen (str) + 1);

  close (fd);

  return 0;
}
/**
 * @author Vinay
 * @date 11.03.2020
 * @brief on usb disconnected state sends usb_disconnected notification via
 * d-bus
 */
static void
on_usb_port_state_disconnected (USBMSG *usbmsg)
{
  gboolean usb_connected = FALSE;
  usbmsgd_debug ("on_usb_port_state_disconnected\n");
  usbmsgd_set_gadget_config_file_content ("");
  check_and_repair_filesystem (usbmsg);

  /* Send the d-bus notification that USB is connected */
  com_bosch_pt_usbmsgd_set_is_usb_connected (usbmsg->usbmsgd_object,
                                             usb_connected);
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief on usb connected state sends usb_connected notification via d-bus
 */
static void
on_usb_port_state_connected (USBMSG *usbmsg)
{

  gboolean usb_connected = TRUE;
  gchar *fstype          = "unknow";
  usbmsgd_debug ("on_usb_port_state_connected\n");

  com_bosch_pt_usbmsgd_set_file_system_type (usbmsg->usbmsgd_object, fstype);
  com_bosch_pt_usbmsgd_set_is_usb_connected (usbmsg->usbmsgd_object,
                                             usb_connected);
}

/**
 * @author Harish, Vinay
 * @date 22.02.2020
 * @brief Handle the usb connection change.
 * @param channel GIOChannel
 * @param f_enCondition GIOCondition
 * @param user_data Data passed during handler installation. By Default NULL.
 */
static gboolean
handle_usb_connection_changed (GIOChannel *channel, GIOCondition f_enCondition,
                               gpointer user_data)
{
  gchar *pgcData;
  gsize gsBytesRead;
  USBMSG *usbmsg = user_data;
  g_io_channel_read_to_end (channel, &pgcData, &gsBytesRead, NULL);

  if (gsBytesRead >= 6)
    {
      if (strncmp (pgcData, "b_idle", 6) == 0)
        {
          on_usb_port_state_disconnected (usbmsg);
        }
      else if (strncmp (pgcData, "b_peri", 6) == 0)
        {
          on_usb_port_state_connected (usbmsg);
        }
      else
        {
          usbmsgd_warning ("Unknown usb port state %s\n", pgcData);
        }
    }
  else
    {
      usbmsgd_warning ("Could not read enough bytes from mode file\n");
    }

  g_io_channel_seek_position (channel, 0, G_SEEK_SET, NULL);

  return TRUE;
}
/**
 * @author Vinay
 * @date 11.03.2020
 * @brief start monitor on usb port for usb connected or disconnected.
 */
static void
start_monitor_usb_port (USBMSG *usbmsg)
{
  if (get_usb_port () && (ms_pc_usb_port != NULL))
    {

      GFile *usb_port_file = g_file_new_for_path (ms_pc_usb_port);

      if (g_file_query_exists (usb_port_file, NULL))
        {
          ms_usb_channel = g_io_channel_new_file (ms_pc_usb_port, "r", NULL);
          if (ms_usb_channel != NULL)
            {
              ms_watch_id
                  = g_io_add_watch (ms_usb_channel, G_IO_PRI,
                                    handle_usb_connection_changed, usbmsg);
            }
          else
            {
              usbmsgd_warning ("Could not start monitor!!!\n");
            }
        }
      else
        {
          usbmsgd_warning (
              "Could not start monitor usb port because sys file related "
              "to port is not found !!!\n");
        }

      g_object_unref (usb_port_file);
    }
  else
    {
      usbmsgd_critical ("USB otg path not found\n");
    }
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief stop monitor on usb port for usb connected or disconnected.
 */
static void
stop_monitor_usb_port ()
{
  g_autoptr (GError) error = NULL;
  if (ms_watch_id)
    {
      g_source_remove (ms_watch_id);
    }
  if (ms_usb_channel != NULL)
    {
      g_io_channel_shutdown (ms_usb_channel, TRUE, &error);
    }
  if (ms_pc_usb_port != NULL)
    {
      g_free (ms_pc_usb_port);
    }
  usbmsgd_debug ("stop_monitor_usb_port done\n");
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief usbmsgd init.
 */
void
usbmsgd_init (USBMSG *usbmsg)
{
  /* initilize the libblock  plugins */
  blockdev_plugin_init ();

  /* Create loop device from backing file */
  setup_loopdev (usbmsg);

  /*
   * Start usb port monitor
   *  This will also get callback if USB already connected or disconnected and
   *  Check filesystem type, check filesystem and repair if check failed
   *  when USB disconnected.
   */
  start_monitor_usb_port (usbmsg);
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief usbmsgd deinit.
 */
void
usbmsgd_deinit (USBMSG *usbmsg)
{
  teardown_loopdev (usbmsg);

  stop_monitor_usb_port ();
}

/**
 * @author Vinay
 * @date 27.02.2020
 * @brief Mount the Mass Storage Partition to Device.
 * @param dbus ComBosch_ptUsbmsgd
 * @param invocation GDBusMethodInvocation - Object for handling remote calls.
 * @param user_data Data passed during handler installation. By Default NULL.
 * @return TRUE on success. FALSE otherwise.
 */
static gboolean
handle_mount (ComBosch_ptUsbmsgd *usb_storage,
              GDBusMethodInvocation *invocation, gpointer user_data)
{
  gint ret                = 0;
  const gchar *mount_path = USBMSGD_MOUNT_DIR;
  gchar *arg_mount_option = NULL;
  USBMSG *usbmsg          = user_data;
  usbmsgd_debug ("handle_mount\n");
  ret = mount_for_device (usbmsg, mount_path, arg_mount_option);
  if (ret)
    {
      com_bosch_pt_usbmsgd_complete_mount_to_device (usb_storage, invocation,
                                                     mount_path);
    }
  else
    {
      com_bosch_pt_usbmsgd_complete_mount_to_device (usb_storage, invocation,
                                                     NULL);
    }
  return TRUE;
}

/**
 * @author Vinay
 * @date 27.02.2020
 * @brief UnMount the Mass Storage Partition.
 * @param usb_storage ComBosch_ptUsbmsgd
 * @param invocation GDBusMethodInvocation - Object for handling remote calls.
 * @param user_data Data passed during handler installation. By Default NULL.
 * @return TRUE on success. FALSE otherwise.
 */
static gboolean
handle_unmount (ComBosch_ptUsbmsgd *usb_storage,
                GDBusMethodInvocation *invocation, gpointer user_data)
{
  usbmsgd_debug ("handle_unmount\n");
  USBMSG *usbmsg = user_data;
  unmount_for_device (usbmsg);
  com_bosch_pt_usbmsgd_complete_unmount_to_device (usb_storage, invocation);
  return TRUE;
}

/**
 * @author Vinay
 * @date 21.08.2020
 * @brief Check the label to be set for the usb mass storage and
 * If label is empty it will set as empty.
 * If label is more than 11 characters it is truncate to 11.
 * If label is lower case letters convert to upper case letters.
 * * @param arg_label label passed by user/client.
 */
static gchar *
check_label_and_convert_to_upper (const gchar *arg_label)
{
  gchar *label;
  if (NULL == arg_label || (0 == strcmp (arg_label, "")))
    {
      label = g_strdup (" ");
      usbmsgd_warning ("label is empty\n");
    }
  else if (strlen (arg_label) > 11)
    {
      usbmsgd_warning ("label is more than 11 characters\n");
      label     = g_strndup (arg_label, 11);
      label[11] = '\0';
    }
  else
    {
      label = g_strdup (arg_label);
    }

  for (int i = 0; label[i] != '\0' && i < 11; i++)
    {
      if (label[i] >= 'a' && label[i] <= 'z')
        {
          label[i] = label[i] - 32; /* Convert to UPPERCASE letter */
        }
    }
  return label;
}

/**
 * @author Vinay
 * @date 11.06.2021
 * @brief freeing the task data
 * @param ptr data pointer.
 */
static void
format_fs_vfat_thread_data_free (gpointer ptr)
{
  format_data *data = ptr;
  free (data->label);
  free (data);
}

/**
 * @author Vinay
 * @date 11.06.2021
 * @brief callback in the current thread-default main context.
 * @param source_object the object the asynchronous operation was started with.
 * @param res a GAsyncResult.
 * @param user_data user data passed to the callback.
 */
static void
format_fs_vfat_thread_finish_cb (GObject *source_object, GAsyncResult *res,
                                 gpointer user_data)
{
  g_autoptr (GError) error          = NULL;
  USBMSG *usbmsg                    = user_data;
  UsbmsgdFormatStatus format_status = USBMSGD_FORMAT_STARTED;

  gssize ret = g_task_propagate_int (G_TASK (res), &error);
  usbmsgd_debug ("Thread Return with ret = %d\n", ret);

  if (ret)
    {
      format_status = USBMSGD_FORMAT_DONE;
      com_bosch_pt_usbmsgd_set_is_file_system_corrupted (
          usbmsg->usbmsgd_object, FALSE);
      /*
       * Is it required to call bd_fs_get_fstype(device_name, &error);
       * before setting property?
       */
      com_bosch_pt_usbmsgd_set_file_system_type (usbmsg->usbmsgd_object,
                                                 "vfat");
    }
  else
    {
      format_status = USBMSGD_FORMAT_FAILED;
    }

  com_bosch_pt_usbmsgd_set_format_status (usbmsg->usbmsgd_object,
                                          format_status);
  g_dbus_interface_skeleton_flush (
      (G_DBUS_INTERFACE_SKELETON (usbmsg->usbmsgd_object)));
}

/**
 * @author Vinay
 * @date 11.06.2021
 * @brief task function to be run in a thread.
 * @param task the GTask
 * @param source task 's source object.
 * @param task_data task 's task data.
 * @param cancellable task 's GCancellable, or NULL.
 */
void
format_fs_vfat_thread (GTask *task, gpointer source, gpointer task_data,
                       GCancellable *cancellable)
{
  g_autoptr (GError) error = NULL;
  gint ret                 = 0;
  gchar *fat_size_32       = "32";
  format_data *data        = task_data;
  USBMSG *usbmsg           = data->usbmsg;

  usbmsgd_debug ("format_fs_vfat_thread STARTED\n");

  if (usbmsg->device_name != NULL && data->label != NULL)
    {
      BDExtraArg bd_label_arg         = { "-n", data->label };
      BDExtraArg bd_fat_arg           = { "-F", fat_size_32 };
      const BDExtraArg *extra_args[3] = { &bd_label_arg, &bd_fat_arg, NULL };
      ret = bd_fs_vfat_mkfs (usbmsg->device_name, extra_args, &error);
      if (!ret)
        {
          usbmsgd_critical ("Error bd_fs_vfat_mkfs: %d\n", ret);
          if (error)
            usbmsgd_critical ("(%s, %d)\n", error->message, error->code);
        }
      else
        {
          usbmsgd_debug ("bd_fs_vfat_mkfs Done: %d\n", ret);
        }
    }
  else
    {
      usbmsgd_critical ("format Failed\n");
    }

  g_task_return_int (task, ret);
}

/**
 * @author Vinay
 * @date 11.06.2021
 * @brief Format the Mass Storage Partition of the Device.
 * @param usb_storage ComBosch_ptUsbmsgd
 * @param invocation GDBusMethodInvocation - Object for handling remote calls.
 * @param arg_label label passed by user/client.
 * @param user_data Data passed during handler installation. By Default NULL.
 * @return TRUE on success. FALSE otherwise.
 * @todo Make it as generic function with filisystem type as argument to
 * support multiple filesystem types.
 */
static gboolean
handle_format_fs_vfat (ComBosch_ptUsbmsgd *usb_storage,
                       GDBusMethodInvocation *invocation,
                       const gchar *arg_label, gpointer user_data)
{
  UsbmsgdFormatStatus format_status = USBMSGD_FORMAT_STARTED;
  format_data *data                 = NULL;
  USBMSG *usbmsg                    = user_data;
  GTask *task                       = NULL;
  usbmsgd_debug ("handle_format_fs_vfat STARTED\n");
  com_bosch_pt_usbmsgd_set_format_status (usbmsg->usbmsgd_object,
                                          format_status);
  /* To send multiple dbus notofication within same method
   * and  in same main loop we have to call flush  */
  g_dbus_interface_skeleton_flush (
      (G_DBUS_INTERFACE_SKELETON (usbmsg->usbmsgd_object)));

  if (usbmsg->device_name != NULL)
    {
      data         = g_new0 (format_data, 1);
      data->usbmsg = usbmsg;
      data->label  = check_label_and_convert_to_upper (arg_label);
      usbmsgd_debug ("Handle:Label to set in Upper case %s\n", data->label);
      task = g_task_new (NULL, NULL, format_fs_vfat_thread_finish_cb,
                         user_data);
      g_task_set_task_data (task, data, format_fs_vfat_thread_data_free);

      /* Run the task in a worker thread and return immediately while that
       * continues in the background. When it’s done it will call @callback in
       * the current thread default main context. */
      g_task_run_in_thread (task, format_fs_vfat_thread);

      g_object_unref (task);
    }
  else
    {
      format_status = USBMSGD_FORMAT_FAILED;
      com_bosch_pt_usbmsgd_set_format_status (usbmsg->usbmsgd_object,
                                              format_status);
    }
  com_bosch_pt_usbmsgd_complete_format_vfat (usb_storage, invocation);
  return TRUE;
}

/**
 * @author Vinay
 * @date 28.05.2020
 * @brief Invoked when a connection to D-bus has been obtained. Installs an
 * interface for an object, that can be remotely accessed via D-bus.
 * @param connection The GDBusConnection to a message bus.
 * @param name       The well known name of the application on the bus.
 * @param user_data  Data that was passed to g_bus_own_name().
 */
void
usbmsgd_bus_acquired (GDBusConnection *connection, const gchar *name,
                      gpointer user_data)
{
  usbmsgd_debug ("Acquired bus: %s\n", name);
  USBMSG *usbmsg = user_data;
  /*
   * Create new object
   * Creates a skeleton (server-) object instace for the D-Bus interface
   * "gdbus-interface-com.bosch_pt.Usbmsgd"
   */
  usbmsg->usbmsgd_object = com_bosch_pt_usbmsgd_skeleton_new ();

  /*
   * Install method/signal handler callbacks
   * Connect the callbacks to signal names for this server object.
   * The handler will be called before the default handler of the singal.
   *               ServerObject | "signal-name (Or) signal-name::detail" |
   *               Callback function |            Data to pass to callback.
   *               Default: NULL.
   */
  g_signal_connect (usbmsg->usbmsgd_object, "handle-unmount-to-device",
                    G_CALLBACK (handle_unmount), usbmsg);
  g_signal_connect (usbmsg->usbmsgd_object, "handle-mount-to-device",
                    G_CALLBACK (handle_mount), usbmsg);
  g_signal_connect (usbmsg->usbmsgd_object, "handle-format-vfat",
                    G_CALLBACK (handle_format_fs_vfat), usbmsg);

  /* Export the objects interface */
  if (!g_dbus_interface_skeleton_export (
          G_DBUS_INTERFACE_SKELETON (usbmsg->usbmsgd_object), connection,
          "/com/bosch_pt/Usbmsgd", NULL))
    {
      usbmsgd_critical (
          "Export ComBosch_ptUsbmsgd, interface skeleton failed\n");
      if (usbmsg->loop)
        g_main_loop_quit (usbmsg->loop);
    }
}

/**
 * @author Vinay
 * @date 28.08.2020
 * @brief Invoked when the name mt.usbmsgd is acquired.
 *
 * @param connection The GDBusConnection on which to acquired to the name.
 * @param name       The name being owned.
 * @param user_data  Passed to g_bus_own_name() or
 * g_bus_own_name_no_connection().
 */
void
usbmsgd_name_acquired (GDBusConnection *connection, const gchar *name,
                       gpointer user_data)
{
  usbmsgd_debug ("Acquired the name: %s\n", name);
  USBMSG *usbmsg = user_data;

  usbmsgd_init (usbmsg);
}

/**
 * @author Vinay
 * @date 28.05.2020
 * @brief Invoked when the name mt.usbmsgd is lost or connection has been
 * closed.
 *
 * @param connection The GDBusConnection on which to acquire the name or NULL
 * if the connection was disconnection.
 * @param name       The name being owned.
 * @param user_data  Passed to g_bus_own_name() or
 * g_bus_own_name_no_connection().
 */
void
usbmsgd_name_lost (GDBusConnection *connection, const gchar *name,
                   gpointer user_data)
{
  usbmsgd_debug ("Lost the name: %s\n", name);
  USBMSG *usbmsg = user_data;
  /* quit GMainLoop. */
  if (usbmsg->loop)
    g_main_loop_quit (usbmsg->loop);
}

/**
 * @author Vinay
 * @date 12.07.2022
 * @brief usbmsgd gadget driver init.
 */
void
usbmsgd_gadget_init ()
{

  usbg_config *c;
  usbg_function *f_ms;
  int ret = -EINVAL;
  int usbg_ret;
  FILE *input;

  struct usbg_f_ms_lun_attrs f_ms_luns_array[] = { {
      .id        = -1, /* allows to place in any position */
      .cdrom     = 0,
      .ro        = 0,
      .nofua     = 0,
      .removable = 1,
      .file      = "/var/lib/mass_storage",
  } };

  struct usbg_f_ms_lun_attrs *f_ms_luns[] = {
    /*
     * When id in lun structure is below 0 we can place it in any
     * arbitrary position
     */
    &f_ms_luns_array[0],
    NULL,
  };

  struct usbg_f_ms_attrs f_attrs = {
    .stall = 0,
    .nluns = 1,
    .luns  = f_ms_luns,
  };

  struct usbg_config_strs c_strs = { "1xMass Storage" };

  /* Prepare input file */
  input = fopen ("/var/lib/gadget_schema.cfg", "r");
  if (!input)
    {
      fprintf (stderr, "Error on fopen. Error: %s\n", strerror (errno));
      fclose (input);
      return;
    }

  usbg_ret = usbg_init ("/sys/kernel/config", &s);
  if (usbg_ret != USBG_SUCCESS)
    {
      usbmsgd_critical ("Error on USB gadget init\n");
      usbmsgd_critical ("Error: %s : %s\n", usbg_error_name (usbg_ret),
                        usbg_strerror (usbg_ret));
      return;
    }
  usbmsgd_warning ("1 usbg_init Done \n");

  usbg_ret = usbg_import_gadget (s, input, "g1", &g);
  if (usbg_ret != USBG_SUCCESS)
    {
      fprintf (stderr, "Error on import gadget\n");
      fprintf (stderr, "Error: %s : %s\n", usbg_error_name (usbg_ret),
               usbg_strerror (usbg_ret));
      if (usbg_ret == USBG_ERROR_INVALID_FORMAT)
        fprintf (stderr, "Line: %d. Error: %s\n",
                 usbg_get_gadget_import_error_line (s),
                 usbg_get_gadget_import_error_text (s));
      return;
    }
  usbmsgd_warning ("2 usbg_import_gadget Done \n");

  usbg_ret
      = usbg_create_function (g, USBG_F_MASS_STORAGE, "0", &f_attrs, &f_ms);
  if (usbg_ret != USBG_SUCCESS)
    {
      usbmsgd_critical ("Error creating mass storage function\n");
      usbmsgd_critical ("Error: %s : %s\n", usbg_error_name (usbg_ret),
                        usbg_strerror (usbg_ret));
      return;
    }
  usbmsgd_warning ("3 usbg_create_function Done \n");

  /* NULL can be passed to use kernel defaults */
  usbg_ret = usbg_create_config (g, 1, "c", NULL, &c_strs, &c);
  if (usbg_ret != USBG_SUCCESS)
    {
      usbmsgd_critical ("Error creating config\n");
      usbmsgd_critical ("Error: %s : %s\n", usbg_error_name (usbg_ret),
                        usbg_strerror (usbg_ret));
      return;
    }
  usbmsgd_warning ("4 usbg_create_config Done \n");

  usbg_ret = usbg_add_config_function (c, "c", f_ms);
  if (usbg_ret != USBG_SUCCESS)
    {
      usbmsgd_critical ("Error adding ms function\n");
      usbmsgd_critical ("Error: %s : %s\n", usbg_error_name (usbg_ret),
                        usbg_strerror (usbg_ret));
      return;
    }
  usbmsgd_warning ("5 usbg_add_config_function Done \n");
  usbg_ret = usbg_enable_gadget (g, DEFAULT_UDC);
  if (usbg_ret != USBG_SUCCESS)
    {
      usbmsgd_critical ("Error enabling gadget\n");
      usbmsgd_critical ("Error: %s : %s\n", usbg_error_name (usbg_ret),
                        usbg_strerror (usbg_ret));
      return;
    }
  usbmsgd_warning ("6 usbg_enable_gadget Done \n");
}

/**
 * @author Vinay
 * @date 12.07.2022
 * @brief usbmsgd gadget driver deinit.
 */
int
usbmsgd_gadget_deinit ()
{
  int usbg_ret;
  usbg_udc *u;

  /* Check if gadget is enabled */
  u = usbg_get_gadget_udc (g);

  /* If gadget is enable we have to disable it first */
  if (u)
    {
      usbg_ret = usbg_disable_gadget (g);
      if (usbg_ret != USBG_SUCCESS)
        {
          usbmsgd_critical ("Error on USB disable gadget udc\n");
          usbmsgd_critical ("Error: %s : %s\n", usbg_error_name (usbg_ret),
                            usbg_strerror (usbg_ret));
          goto out;
        }
    }
  usbmsgd_warning ("7 usbg_disable_gadget Done \n");

  /* Remove gadget with USBG_RM_RECURSE flag to remove
   * also its configurations, functions and strings */
  usbg_ret = usbg_rm_gadget (g, USBG_RM_RECURSE);
  if (usbg_ret != USBG_SUCCESS)
    {
      usbmsgd_critical ("Error on USB gadget remove\n");
      usbmsgd_critical ("Error: %s : %s\n", usbg_error_name (usbg_ret),
                        usbg_strerror (usbg_ret));
    }
  usbmsgd_warning ("8 usbg_rm_gadget Done \n");

out:
  usbg_cleanup (s);
  usbmsgd_warning ("9 usbg_cleanup Done \n");
  return usbg_ret;
}
